#include "AbstractDiagnostic.h"

#ifdef Q_OS_LINUX

#elif defined(Q_OS_WIN)
#include "registry.h"
#elif defined(Q_OS_MACOS)


#endif

AbstractDiagnostic::AbstractDiagnostic(QObject *parent)
    :QObject(parent)
{
#ifdef Q_OS_LINUX
    s_nodeDataPath = "/opt/cellframe-node";
    s_sendedDataFilePath = "/tmp/lastSendedData.json";
#elif defined(Q_OS_WIN)
    s_nodeDataPath = QString("%1/cellframe-node").arg(regGetUsrPath());
    s_sendedDataFilePath = QString("%1/cellframe-node/lastSendedDiagData.json").arg(regGetUsrPath());
#elif defined(Q_OS_MACOS)
    s_nodeDataPath = QString("Applications/CellframeNode.app/Contents/Resources/");
    s_sendedDataFilePath = "/tmp/lastSendedData.json";
#endif

    s_jsonFilePath = QString("%1/etc/diagdata.json").arg(s_nodeDataPath);

    QFile config_file(s_jsonFilePath);
    if(!config_file.exists())
    {
        config_file.open(QIODevice::WriteOnly);
        QJsonObject obj{{"category","other"}};
        config_file.write(QJsonDocument(obj).toJson());
        config_file.close();
    }

    s_timer_update = new QTimer();
    s_mac = get_mac();
}

AbstractDiagnostic::~AbstractDiagnostic()
{
    delete s_timer_update;
}

void AbstractDiagnostic::set_timeout(int timeout){
    s_timer_update->stop();
    s_timeout = timeout;
    s_timer_update->start(s_timeout);
}

void AbstractDiagnostic::start_diagnostic()
{
    s_timer_update->start(s_timeout);
}

void AbstractDiagnostic::stop_diagnostic()
{
    s_timer_update->stop();
}

QJsonValue AbstractDiagnostic::get_mac()
{
    QString MAC{"unknown"};
    foreach(QNetworkInterface netInterface, QNetworkInterface::allInterfaces())
    {
        // Return only the first non-loopback MAC Address
        if (!(netInterface.flags() & QNetworkInterface::IsLoopBack))
        {
            qDebug()<<netInterface.hardwareAddress();
            if(!netInterface.hardwareAddress().isEmpty())
            {
                MAC = netInterface.hardwareAddress();
                break;
            }
        }
    }

    return MAC;
}

QString AbstractDiagnostic::get_uptime_string(long sec)
{
    QTime time(0, 0);
    time = time.addSecs(sec);
    int fullHours = sec/3600;

    QString uptime = QString("%1:").arg(fullHours) + time.toString("mm:ss");

    return uptime;
}

quint64 AbstractDiagnostic::get_file_size (QString flag, QString path ) {

    if(flag == "log")
        path += "/var/log";
    else
    if (flag == "DB")
        path += "/var/lib/global_db";
    else
    if (flag == "chain")
        path += "/var/lib/network";
    else
        path += "";

    QDir currentFolder( path );

    quint64 totalsize = 0;

    currentFolder.setFilter( QDir::Dirs | QDir::Files | QDir::NoSymLinks );
    currentFolder.setSorting( QDir::Name );

    QFileInfoList folderitems( currentFolder.entryInfoList() );

    foreach ( QFileInfo i, folderitems ) {
        QString iname( i.fileName() );
        if ( iname == "." || iname == ".." || iname.isEmpty() )
            continue;
        if(flag == "log" && i.suffix() != "log" && !i.isDir())
            continue;
        else
        if(flag == "DB" && (i.suffix() != "dat" && !i.suffix().isEmpty()) && !i.isDir())
            continue;
        else
        if(flag == "chain" && i.suffix() != "dchaincell" && !i.isDir())
            continue;

        if ( i.isDir() )
            totalsize += get_file_size("", path+"/"+iname);
        else
            totalsize += i.size();
    }

    return totalsize;
}

QString AbstractDiagnostic::get_memory_string(long num)
{
    QString result = QString::number(num);
    return result;

    int gb = (num / 1024) / 1024;
    int mb = (num-gb*1024*1024) /1024;
    int kb = (num - (gb*1024*1024+mb*1024));
    if (gb > 0)
       result = QString::number(gb) + QString(" Gb ");
    else
       result = QString("");
    if (mb > 0)
       result += QString::number(mb) + QString(" Mb ");
    if (kb > 0)
       result += QString::number(kb) + QString(" Kb ");

    return result;
}

QJsonObject AbstractDiagnostic::roles_processing()
{
    QJsonObject rolesObject;

    QDir currentFolder(s_nodeDataPath + "/etc/network");

    currentFolder.setFilter( QDir::Dirs | QDir::Files | QDir::NoSymLinks );
    currentFolder.setSorting( QDir::Name );

    QFileInfoList folderitems( currentFolder.entryInfoList() );

    foreach ( QFileInfo i, folderitems ) {
        QString iname( i.fileName() );
        if ( iname == "." || iname == ".." || iname.isEmpty() )
            continue;
        if(i.suffix() == "cfg" && !i.isDir())
        {
            QSettings config(i.absoluteFilePath(), QSettings::IniFormat);

            rolesObject.insert(i.completeBaseName(), config.value("node-role", "unknown").toString());
        }
    }

    return rolesObject;
}

/// ---------------------------------------------------------------
///        Cli info
/// ---------------------------------------------------------------
QJsonObject AbstractDiagnostic::get_cli_info()
{
    QStringList networks = get_networks();

    QJsonObject netObj;

    for(QString net : networks)
    {
        QJsonObject dataObj;

        dataObj.insert("net_info", get_net_info(net));
        dataObj.insert("mempool" , get_mempool_count(net));
//        dataObj.insert("ledger"  , get_ledger_count(net));
        dataObj.insert("blocks"  , get_blocks_count(net));
        dataObj.insert("events"  , get_events_count(net));
        dataObj.insert("nodelist"  , get_nodelist(net));

        netObj.insert(net, dataObj);
    }

    return netObj;
}

QStringList AbstractDiagnostic::get_networks()
{
    QProcess proc;
    proc.start(QString(CLI_PATH), QStringList()<<"net"<<"list");
    proc.waitForFinished(5000);
    QString result = proc.readAll();

    QStringList listNetworks;
    result.remove(' ');
    result.remove("\r");
    result.remove("\n");
    result.remove("networks:");
    if(!(result.isEmpty() || result.isNull() || result.contains('\'') || result.contains("error") || result.contains("Error") || result.contains("err")))
    {
        listNetworks = result.split(",", QString::SkipEmptyParts);
    }
    
    return listNetworks;
}

QJsonObject AbstractDiagnostic::get_net_info(QString net)
{
    QProcess proc;
    proc.start(QString(CLI_PATH),
               QStringList()<<"net"<<"-net"<<QString(net)<<"get"<<"status");
    proc.waitForFinished(5000);
    
    QString result = proc.readAll();
    
    // ---------- Status ----------------
    QRegularExpression rx_state(R"(.*current: ([A-Z,_]+).*)");
    QRegularExpression rx_addr(R"(.*current_addr: (.+))");
    
    QJsonObject resultObj;
    
    QRegularExpressionMatch match_addr = rx_addr.match(result);
    if (match_addr.hasMatch()) {
        resultObj.insert("node_address", match_addr.captured(1));
    }

    QRegularExpressionMatch match_state = rx_state.match(result);
    if (match_state.hasMatch()) {
        resultObj.insert("state", match_state.captured(1));
    }

    // ---------- Links count ----------------
    QRegularExpression rxLinksActive(R"(.*active: (.*))");
    QRegularExpression rxLinksRequired(R"(.*required: (.*))");
    
    QRegularExpressionMatch match_active = rxLinksActive.match(result);
    if (match_active.hasMatch()) {
        resultObj.insert("active_links_count", match_active.captured(1));
    }
    
    QRegularExpressionMatch match_req = rxLinksRequired.match(result);
    if (match_req.hasMatch()) {
        resultObj.insert("links_count", match_req.captured(1));
    }
    
    // ---------- Sync status ----------------
    QRegularExpression rxZeroChainSync(R"(zerochain: \s+status: (\S+|.+)\s+current: (\S+)\s+in network: (\S+)\s+percent: (\S+|.+))");
    QJsonObject objZeroChain;

    QRegularExpressionMatch match_zeroChain = rxZeroChainSync.match(result);
    if (match_zeroChain.hasMatch()) {
        objZeroChain.insert("status",  match_zeroChain.captured(1));
        objZeroChain.insert("current", match_zeroChain.captured(2));
        objZeroChain.insert("network", match_zeroChain.captured(3));
        objZeroChain.insert("percent", match_zeroChain.captured(4));
    }

    QRegularExpression rxMainChainSync(R"(main: \s+status: (\S+|.+)\s+current: (\S+)\s+in network: (\S+)\s+percent: (\S+|.+))");
    QJsonObject objMainChain;

    QRegularExpressionMatch match_mainChain = rxMainChainSync.match(result);
    if (match_mainChain.hasMatch()) {
       objMainChain.insert("status",  match_mainChain.captured(1));
       objMainChain.insert("current", match_mainChain.captured(2));
       objMainChain.insert("network", match_mainChain.captured(3));
       objMainChain.insert("percent", match_mainChain.captured(4));
    }

    QJsonObject syncObj{
       {"main", objMainChain},
       {"zero", objZeroChain}
    };

    resultObj.insert("sync_status", syncObj);
    resultObj.insert("balancer", get_balancer_links(net));

    return resultObj;
}

QJsonObject AbstractDiagnostic::get_mempool_count(QString net)
{
    QProcess proc;
    proc.start(QString(CLI_PATH),
               QStringList()<<"mempool"<<"count"<<"-net"<<QString(net));
    proc.waitForFinished(5000);
    QString result = proc.readAll();
    
    QRegularExpression rx(R"(name: (.*)\n.*count: (\d+))");

    QJsonObject resultObj;

    QRegularExpressionMatchIterator matchItr = rx.globalMatch(result);

    while (matchItr.hasNext())
    {
        
        QRegularExpressionMatch match = matchItr.next();
        resultObj.insert(match.captured(1), match.captured(2));
    }
    return resultObj;
}

QJsonObject AbstractDiagnostic::get_ledger_count(QString net)
{
    //TODO: legder tx -all -net   NOT WORKING
    return {};
}

QJsonObject AbstractDiagnostic::get_blocks_count(QString net)
{
    QProcess proc;
    proc.start(QString(CLI_PATH),
               QStringList()<<"block"<<"last"<<"-net"<<QString(net) <<"-chain" << "main");
    proc.waitForFinished(5000);
    QString result = proc.readAll();
    
    QRegularExpression rx_block_count(R"(\.(.+) has blocks: (.+))");
    QRegularExpression rx_creation(R"(.*ts_created: (.*))");
    QRegularExpression rx_hash(R"(.*Last block hash: (.*))");
    

    QRegularExpressionMatch block_count_match = rx_block_count.match(result);
    QRegularExpressionMatch timestamp_match = rx_creation.match(result);
    QRegularExpressionMatch hash_match = rx_hash.match(result);

    if (!block_count_match.hasMatch()) {
        return {};
    }

    QJsonObject resultObj;
    resultObj.insert(block_count_match.captured(1), block_count_match.captured(2));

    QJsonObject last_block;
    last_block.insert("hash", hash_match.captured(1));
    
    qDebug() << "TS: "<<timestamp_match.captured(1);
    QDateTime dt = QDateTime::fromString(timestamp_match.captured(1), Qt::RFC2822Date);
    qint64 timestamp = dt.toSecsSinceEpoch();
    last_block.insert("timestamp", QString::number(timestamp));

    resultObj.insert("last_block", last_block);

    qDebug() << "Block last "<<net<<":";
    qDebug() << resultObj;
    return resultObj;

}

QJsonObject AbstractDiagnostic::get_events_count(QString net)
{
    QProcess proc;
    proc.start(QString(CLI_PATH),
               QStringList()<<"dag"<<"event"<<"last"<<"-net"<<QString(net) <<"-chain" << "zerochain");
    proc.waitForFinished(5000);
    QString result = proc.readAll();

    QRegularExpression rx_event_count(R"(\.(.+) has events: (.+))");
    QRegularExpression rx_creation(R"(.*ts_created: (.*))");
    QRegularExpression rx_hash(R"(.*Last event hash: (.*))");
    

    QRegularExpressionMatch event_count_match = rx_event_count.match(result);
    QRegularExpressionMatch timestamp_match = rx_creation.match(result);
    QRegularExpressionMatch hash_match = rx_hash.match(result);

    if (!event_count_match.hasMatch()) {
        return {};
    }

    QJsonObject resultObj;
    resultObj.insert(event_count_match.captured(1), event_count_match.captured(2));

    QJsonObject last_event;
    last_event.insert("hash", hash_match.captured(1));
    
    qDebug() << "TS: "<<timestamp_match.captured(1);
    QDateTime dt = QDateTime::fromString(timestamp_match.captured(1), Qt::RFC2822Date);
    qint64 timestamp = dt.toSecsSinceEpoch();
    last_event.insert("timestamp", QString::number(timestamp));

    resultObj.insert("last_event", last_event);

    last_event.insert("timestamp", QString::number(timestamp));

    qDebug() << "Event last "<<net<<":";
    qDebug() << resultObj;
    
    return resultObj;
}


QJsonArray AbstractDiagnostic::get_nodelist(QString net)
{
    QProcess proc;
    proc.start(QString(CLI_PATH),
               QStringList()<<"node"<<"dump"<<"-net"<<QString(net));
    proc.waitForFinished(5000);
    QString result = proc.readAll();

    QJsonArray res;
    for (auto l : result.split("\n"))
        res.push_back(QJsonValue(l));

    return res;
}

QJsonObject AbstractDiagnostic::get_balancer_links(QString net)
{
    QProcess proc;
    proc.start(QString(CLI_PATH),
                QStringList()<<"node"<<"connections"<<"-net"<<QString(net));


    proc.waitForFinished(5000);
    QString result = proc.readAll();

    QJsonObject resultObj;

    QRegularExpression rx(R"(Total links: \d+ \| Uplinks: (\d+) \| Downlinks: (\d+))");
    QRegularExpressionMatch match = rx.match(result);

    if (!match.hasMatch()) {
        return {};
    }

    resultObj.insert("uplinks", match.captured(1));
    resultObj.insert("downlinks", match.captured(2));


    return resultObj;
}
