#include "NotifyWorker.h"
#include "NodeTrayCommandController.h"

QByteArrayList NotifyWorker::jsonListFromData(QByteArray data)
{
    return data.split('\x00');
}

NotifyWorker::NotifyWorker(QObject *parent)
    :QObject(parent)
{
    NodeTrayCommandController *m_trayCtrl = &NodeTrayCommandController::getInstance();

    if(m_trayCtrl->initConfTool())
    {
        m_configWorkerPath = m_trayCtrl->m_confToolPath;
        m_initTimer = new QTimer(this);
        m_reconnectTimer = new QTimer(this);

        if(initWatcher())
            return;
        else
        {
            m_initTimer->start(5000);
            connect(m_initTimer, &QTimer::timeout, [=] {
                qDebug()<<"Reinit timer tick";
                if(initWatcher())
                    m_initTimer->stop();
            });
        }
    }
    m_statusInitWatcher = false;
}

NotifyWorker::~NotifyWorker()
{
    delete m_initTimer;
    delete m_reconnectTimer;
}

QVariant NotifyWorker::readConfig(QString filename, QString section, QString param, QVariant defaultValue)
{
    QProcess proc;
    proc.setProgram(m_configWorkerPath);
    proc.setArguments(QStringList()<<"-e"<<"config"<<filename<<section<<param<<"get");

    proc.start();

    if(proc.waitForFinished())
    {
        QByteArray res = proc.readAllStandardOutput();

        QString strRes = QString(res).remove("\n").remove("\r").remove("\t");
        return strRes.split("=")[1];
    }

    qDebug()<<proc.readAllStandardError();
    return defaultValue;
}

bool NotifyWorker::initWatcher()
{
    m_listenPath = readConfig("cellframe-node", "notify_server", "listen_path", "").toString();
    m_listenAddr = readConfig("cellframe-node", "notify_server", "listen_address", "").toString();
    m_listenPort = readConfig("cellframe-node", "notify_server", "listen_port", 0).toUInt();

    if(m_listenAddr.contains(":"))
    {
        m_listenAddr = m_listenAddr.contains("[") ? m_listenAddr.remove("[") : m_listenAddr;
        m_listenAddr = m_listenAddr.contains("]") ? m_listenAddr.remove("]") : m_listenAddr;

        m_listenPort = QString(m_listenAddr.split(":")[1]).toUInt();
        m_listenAddr = m_listenAddr.split(":")[0];
    }

    qDebug() << "Tcp config: " << m_listenAddr << m_listenPort;
    connect(m_reconnectTimer, SIGNAL(timeout()), this, SLOT(slotReconnect()));

    if (!m_listenPath.isEmpty())
    {
        m_socket = new QLocalSocket(this);
        connect((QLocalSocket*)m_socket, QOverload<QLocalSocket::LocalSocketError>::of(&QLocalSocket::error),
                this, &NotifyWorker::slotError);

        connect((QLocalSocket*)m_socket, &QLocalSocket::connected,
                this, &NotifyWorker::socketConnected);

        connect((QLocalSocket*)m_socket, &QLocalSocket::disconnected,
                this, &NotifyWorker::socketDisconnected);

        connect((QAbstractSocket*)m_socket, &QAbstractSocket::stateChanged,
                this, &NotifyWorker::socketStateChanged);

        connect(m_socket, &QLocalSocket::readyRead, this, &NotifyWorker::socketReadyRead);

        ((QLocalSocket*)m_socket)->connectToServer(m_listenPath);
        ((QLocalSocket*)m_socket)->waitForConnected();
    }
    else
    {
        m_socket = new QTcpSocket(this);
        connect((QTcpSocket*)m_socket, QOverload<QAbstractSocket::SocketError>::of(&QAbstractSocket::error),
                this, &NotifyWorker::slotError);

        connect((QTcpSocket*)m_socket, &QTcpSocket::connected,
                this, &NotifyWorker::socketConnected);

        connect((QTcpSocket*)m_socket, &QTcpSocket::disconnected,
                this, &NotifyWorker::socketDisconnected);

        connect((QTcpSocket*)m_socket, &QTcpSocket::stateChanged,
                this, &NotifyWorker::socketStateChanged);

        connect(m_socket, &QTcpSocket::readyRead, this, &NotifyWorker::socketReadyRead);

        ((QTcpSocket*)m_socket)->connectToHost(m_listenAddr, m_listenPort);
        ((QTcpSocket*)m_socket)->waitForConnected();
    }
    m_statusInitWatcher = true;
    return true;
}


void NotifyWorker::slotError()
{
    qWarning() << "Notify socket error" << m_socket->errorString();
    reconnectFunc();
}

void NotifyWorker::slotReconnect()
{
    qInfo()<<"NotifyWorker::slotReconnect()" << m_listenAddr << m_listenPort << "Socket state" << m_socketState;

    ((QTcpSocket*)m_socket)->connectToHost(m_listenAddr, m_listenPort);
    ((QTcpSocket*)m_socket)->waitForConnected(5000);
}

void NotifyWorker::socketConnected()
{
    qInfo() << "Notify socket connected";
    m_reconnectTimer->stop();
    m_socket->waitForReadyRead(4000);
}

void NotifyWorker::socketDisconnected()
{
    qWarning() << "Notify socket disconnected";
    reconnectFunc();
}

void NotifyWorker::socketStateChanged(QAbstractSocket::SocketState socketState)
{
    qDebug() << "Socket state changed" << socketState;
    m_socketState = socketState;
}

void NotifyWorker::socketReadyRead()
{
    //    qDebug() << "Ready Read";
    QByteArray data = m_socket->readLine();
    //    QByteArray data = socket->readAll();
    if (data[data.length() - 1] != '}')
        data = data.left(data.length() - 1);

    if (data[0] != '{')
        data = data.right(data.length() - 1);

    QByteArrayList list = jsonListFromData(data);

    for (int i = 0; i < list.length(); ++i)
    {
        QJsonParseError error;
        QJsonDocument reply = QJsonDocument::fromJson(list[i], &error);
        if (error.error != QJsonParseError::NoError) {
            //            qWarning()<<"Notify parse error. " << error.errorString(); // more logs
            continue;
        }

        if(reply.object()["class"].toString() == "chain_init")
            procNotifyData(reply.object());
    }
}

void NotifyWorker::procNotifyData(const QJsonObject &rcvObj)
{
    m_rcvInitProgress = true;

    bool isContains{false};
    int itr{0};

    for (const auto &net: qAsConst(m_arrNodeInit)) {
        const auto findNet = net.toObject();

        if (findNet["name"] == rcvObj["net"]) {
            isContains = true;

            QJsonObject objChains = m_arrNodeInit[itr].toObject()["chains"].toObject();

            if(rcvObj["chain_id"].toInt() == 1)
            {
                objChains["main"] = rcvObj["load_progress"];
                objChains["zero"] = 100;
            }
            else
                objChains["zero"] = rcvObj["load_progress"];

            QJsonObject object
            {
                {"name", rcvObj["net"].toString()},
                {"chains", objChains}
            };

            m_arrNodeInit[itr] = object;
            break;
        }
        itr++;
    }

    QJsonObject result{
        {"init_progress", m_arrNodeInit},
        {"notify_status", m_socketState == QAbstractSocket::SocketState::ConnectedState?
                                           "connected" : "disconnected"}
    };

    m_docNodeInit.setObject(result);

//    qDebug()<<m_docNodeInit.toJson();
}

void NotifyWorker::nodeIsOnline()
{
    if(!m_rcvInitProgress)
    {
        m_arrNodeInit = QJsonArray();
        for(const auto &net: qAsConst(m_netList))
        {
            QJsonObject chains;
            chains.insert("main", 100);
            chains.insert("zero", 100);

            QJsonObject obj;
            obj.insert("name", net);
            obj.insert("chains", chains);

            m_arrNodeInit.append(obj);
        }

        QJsonObject result{
            {"init_progress", m_arrNodeInit},
            {"notify_status", m_socketState == QAbstractSocket::SocketState::ConnectedState?
                                  "connected" : "disconnected"}
        };

        m_docNodeInit.setObject(result);
    }
}

void NotifyWorker::rcvNetObj(QJsonObject objNet)
{
    m_netList = objNet.keys();

    for(const auto &net: qAsConst(m_netList))
    {
        QJsonObject chains;
        chains.insert("zero", 0);
        chains.insert("main", 0);


        QJsonObject obj;
        obj.insert("name", net);
        obj.insert("chains", chains);

        m_arrNodeInit.append(obj);
    }
}

void NotifyWorker::reconnectFunc()
{
    m_reconnectTimer->stop();

    if(m_socketState != QAbstractSocket::SocketState::ConnectedState &&
        m_socketState != QAbstractSocket::SocketState::ConnectingState)
    {
        m_reconnectTimer->start(10000);
        qWarning()<< "Notify socket reconnecting...";
    }
}
