#include "NodeTrayCommandController.h"

NodeTrayCommandController::NodeTrayCommandController(QObject *parent)
    : QObject{parent}
{

}

NodeTrayCommandController &NodeTrayCommandController::getInstance()
{
    static NodeTrayCommandController instance;
    return instance;
}

void NodeTrayCommandController::init()
{
    m_configDir = getConfigPath();
    m_statusInitConftool = initConfTool();

    initEngine();
}

void NodeTrayCommandController::initEngine()
{
#if defined(Q_OS_LINUX) && !defined(Q_OS_ANDROID)
    QString os = "linux";
#elif defined Q_OS_WIN
    QString os = "win";
#elif defined Q_OS_MAC
    QString os = "macos";
#else
    QString os = "unknown";
#endif

    QQmlContext * context = m_engine.rootContext();
    context->setContextProperty("trayCommandController", this);
    context->setContextProperty("diagnosticWorker", &diagnosticWorker);
    context->setContextProperty("CURRENT_OS", QVariant::fromValue(os));

    const QUrl url(QStringLiteral("qrc:/main.qml"));
    QObject::connect(&m_engine, &QQmlApplicationEngine::objectCreated,
        this, [url](QObject *obj, const QUrl &objUrl) {
            if (!obj && url == objUrl)
                QCoreApplication::exit(-1);
        }, Qt::QueuedConnection);

    m_engine.load(url);
}


QQmlApplicationEngine *NodeTrayCommandController::qmlEngine()
{
    return &m_engine;
}

bool NodeTrayCommandController::initConfTool()
{
    if(!QDir(m_configDir).exists()){
        showMessage("Could not find config path");
    }

    QString pathConftool;
#if defined(Q_OS_LINUX) && !defined(Q_OS_ANDROID)
    pathConftool = "/opt/cellframe-node/bin/cellframe-node-config";
#elif defined (Q_OS_MACOS)
    pathConftool = "/Applications/CellframeNode.app/Contents/MacOS/cellframe-node-config";
#elif defined (Q_OS_WIN)
    //HKLM "Software\${APP_NAME}" "Path"
    HKEY hKey;
    LONG lRes = RegOpenKeyExW(HKEY_LOCAL_MACHINE, L"SOFTWARE\\cellframe-node\\", 0, KEY_READ, &hKey);
    bool bExistsAndSuccess (lRes == ERROR_SUCCESS);
    bool bDoesNotExistsSpecifically (lRes == ERROR_FILE_NOT_FOUND);
    std::wstring path;
    GetStringRegKey(hKey, L"Path", path, L"");
    std::string stdpath(path.begin(),path.end());

    pathConftool = QString(QString::fromWCharArray(path.c_str()) + "/cellframe-node-config.exe");
#endif

    QFileInfo confTool(pathConftool);
    if(!confTool.exists())
    {
        qWarning()<<"Could not find cellframe-node-config";
        showMessage();
        return false;
    }
    else
    {
        m_confToolPath = confTool.absoluteFilePath();

        return true;
    }
}

void NodeTrayCommandController::runCli()
{
    if(!m_statusInitConftool)
        showMessage();

#if defined(Q_OS_LINUX) && !defined(Q_OS_ANDROID)

    QProcess consoleCli;
    QString cliPath;
    cliPath="/opt/cellframe-node/bin/cellframe-node-cli";
    consoleCli.start("xfce4-terminal", QStringList()<< "--geometry=70x20" << "--command" << cliPath);
    consoleCli.waitForFinished(-1);
#elif defined (Q_OS_MACOS)

#elif defined (Q_OS_WIN)

#endif
}

void NodeTrayCommandController::runTool()
{
    if(!m_statusInitConftool)
        showMessage();

#if defined(Q_OS_LINUX) && !defined(Q_OS_ANDROID)

    QProcess consoleCli;
    QString cliPath;
    cliPath="/opt/cellframe-node/bin/cellframe-node-tool";
    consoleCli.start("xfce4-terminal", QStringList()<< "--geometry=70x20" << "--command" << cliPath);
    consoleCli.waitForFinished(-1);
#elif defined (Q_OS_MACOS)

#elif defined (Q_OS_WIN)

#endif
}

void NodeTrayCommandController::openDir(const QString &dirName)
{
    if(!m_statusInitConftool)
        showMessage();

    QString stringPath = m_configDir;

    QDir dir(QString(stringPath+dirName));
    QDesktopServices::openUrl(QUrl(QUrl::fromLocalFile(dir.absolutePath())));
}

QString NodeTrayCommandController::getResult(QString find, QStringList list)
{
    for(int i = 0; i < list.length(); i++)
    {
        if(list[i].contains(find))
            return list[i].remove(find);
    }
    return "";
}

void NodeTrayCommandController::serviceCommand(TypeServiceCommands type)
{
    if(!m_statusInitConftool)
        showMessage();

    qInfo()<<type;

    switch (type) {
    case Status:
    {
        QString res = sendRequest("-e service status");
        QStringList resList = res.split("\n");

        bool procStatus = getResult("process: ", resList).contains("running");
        bool serviceStatus = getResult("service: ", resList).contains("enabled");

        sigResultStatus(serviceStatus);
//        showMessage("cellframe-node running: " + procStatus);
        qDebug()<<res;
        break;
    }
    case EnableAutostart:
    {
        QString res = sendRequest("-e service enable");
        bool status = !res.contains("disabled");
        sigResultStatus(status);
        showMessage("cellframe-node service: " + res);
        qDebug()<<res;
        break;
    }
    case DisableAutostart:
    {
        QString res = sendRequest("-e service disable");
        bool status = !res.contains("disabled");
        sigResultStatus(status);
        showMessage("cellframe-node service: " + res);
        qDebug()<< res;
        break;
    }
    case Stop:
    {
        QString res = sendRequest("-e service stop");
        showMessage("cellframe-node: " + res);
        qDebug()<< res;
        break;
    }
    case Start:
    {
        QString res = sendRequest("-e service start");
        showMessage("cellframe-node: " + res);
        qDebug()<< res;
        break;
    }
    case Restart:
    {
        QString res = sendRequest("-e service restart");
        showMessage("cellframe-node: " + res);
        qDebug()<< res;
        break;
    }
    default:
        break;
    }
}

QString NodeTrayCommandController::sendRequest(QString req)
{
    QProcess proc;
    proc.setProgram(m_confToolPath);
    proc.setArguments(req.split(" "));

    proc.start();

    if(proc.waitForFinished())
    {
        return proc.readAll();
    }
    else
    {
        showMessage("Timeout request");
        return QString();
    }
}

QString NodeTrayCommandController::getConfigPath()
{
#ifdef __linux__
    return "/opt/cellframe-node/";
#endif

#ifdef __APPLE__
    return  "/Applications/CellframeNode.app/Contents/Resources/";
#endif

#ifdef WIN32
    HKEY hKey;
    LONG lRes = RegOpenKeyExW(HKEY_LOCAL_MACHINE, L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders", 0, KEY_READ, &hKey);
    bool bExistsAndSuccess (lRes == ERROR_SUCCESS);
    bool bDoesNotExistsSpecifically (lRes == ERROR_FILE_NOT_FOUND);
    std::wstring path;
    GetStringRegKey(hKey, L"Common Documents", path, L"");
    std::string stdpath(path.begin(),path.end());
    return QString(QString::fromWCharArray(path.c_str()) + "/cellframe-node");
#endif
}


void NodeTrayCommandController::showMessage(const QString msg, const QMessageBox::Icon icn)
{
//    QMessageBox msgBox;
//    msgBox.setText(msg);
//    msgBox.setIcon(icn);
//    msgBox.exec();

    emit sigShowMessage("Cellframe Node Tray",msg);
}

#ifdef WIN32
LONG NodeTrayCommandController::GetDWORDRegKey(HKEY hKey, const std::wstring &strValueName, DWORD &nValue, DWORD nDefaultValue)
{
    nValue = nDefaultValue;
    DWORD dwBufferSize(sizeof(DWORD));
    DWORD nResult(0);
    LONG nError = ::RegQueryValueExW(hKey,
                                     strValueName.c_str(),
                                     0,
                                     NULL,
                                     reinterpret_cast<LPBYTE>(&nResult),
                                     &dwBufferSize);
    if (ERROR_SUCCESS == nError)
    {
        nValue = nResult;
    }
    return nError;
}


LONG NodeTrayCommandController::GetBoolRegKey(HKEY hKey, const std::wstring &strValueName, bool &bValue, bool bDefaultValue)
{
    DWORD nDefValue((bDefaultValue) ? 1 : 0);
    DWORD nResult(nDefValue);
    LONG nError = GetDWORDRegKey(hKey, strValueName.c_str(), nResult, nDefValue);
    if (ERROR_SUCCESS == nError)
    {
        bValue = (nResult != 0) ? true : false;
    }
    return nError;
}


LONG NodeTrayCommandController::GetStringRegKey(HKEY hKey, const std::wstring &strValueName, std::wstring &strValue, const std::wstring &strDefaultValue)
{
    strValue = strDefaultValue;
    WCHAR szBuffer[512];
    DWORD dwBufferSize = sizeof(szBuffer);
    ULONG nError;
    nError = RegQueryValueExW(hKey, strValueName.c_str(), 0, NULL, (LPBYTE)szBuffer, &dwBufferSize);
    if (ERROR_SUCCESS == nError)
    {
        strValue = szBuffer;
    }
    return nError;
}
#endif
